/******************************************************************************************

   English16.java - Sam Loyd's "Fore and Aft" puzzle
   Version 2.1
   www.mazeworks.com
   
   Copyright (c) 2002 Robert Kirkland
   All Rights Reserved.

   Permission is hereby granted, free of charge, to any person obtaining a copy 
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, and/or sell copies
   of the Software, and to permit persons to whom the Software is furnished to
   do so, provided that the above copyright notice and this permission notice
   appear in all copies of the Software and that both the above copyright
   notice and this permission notice appear in supporting documentation.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER INCLUDED IN THIS NOTICE BE LIABLE FOR
   ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Except as contained in this notice, the name of the copyright holder shall
   not be used in advertising or otherwise to promote the sale, use or other
   dealings in this Software without prior written authorization of the
   copyright holder.

******************************************************************************************/

import java.applet.* ;
import java.awt.* ;
import java.awt.image.*;
import java.net.* ;
import java.util.* ;
	
/******** ENGLISH16 ********
   applet startup - paints Load message, starts main class
*/   
public class English16 extends Applet {
   // applet info
   public String getAppletInfo() { 
      return "English 16 v.2.1 - Copyright (c) 2002 Robert Kirkland"; }
   String parameterInfo[][] = {
      {"bgColor",    "integer", "Background color (24-bit RGB hex int)"},
      {"boardFile",  "string",  "Board image file"},
      {"piecesFile", "string",  "Pieces image file"},
      {"title",      "string",  "Applet title text"},
      {"restart",    "string",  "Restart button text"},
      {"undo",       "string",  "Undo button text"},
      {"moves",      "string",  "Moves label text"},
      {"solve",      "string",  "Solve button text"},
      {"ready",      "string",  "Ready status text"},
      {"win",        "string",  "Win status text"},
      {"lose",       "string",  "Lose status text"},
      {"perfect",    "string",  "Perfect game status text"},
      {"searching",  "string",  "Searching status text"},
      {"found",      "string",  "Solution Found status text"},
      {"solving",    "string",  "Solving status text"},
      {"solved",     "string",  "Solved status text"} } ;
   public String[][] getParameterInfo() { return parameterInfo ; }

   static final Color DF_BG_COLOR = Color.white ;    // default background color
   protected Color BG_COLOR ;
   static final String LOAD_MSG = "Loading program . . ." ;
   protected int appWidth, appHeight ;
   private boolean startup=true, ready=false ;
   
   public void init() {
      // get startup parameters
      appWidth = Integer.parseInt(getParameter("WIDTH")) ;
      appHeight = Integer.parseInt(getParameter("HEIGHT")) ;
      // background color
      String bgColorString = getParameter("bgColor") ;
      try { BG_COLOR = bgColorString==null ? DF_BG_COLOR : new Color(Integer.parseInt(bgColorString,16)) ; } 
      catch (NumberFormatException e) {
         BG_COLOR = DF_BG_COLOR ;
         System.err.println("ERROR: Invalid bgColor parameter.") ;
      }
   }
   private void initialize() {
      AppMain main = new AppMain(this) ;
      main.initialize() ;
   }
   public void paint(Graphics g) {
      drawLoadMsg(g) ;
      if (startup) { 
         startup = false ;
         initialize() ;
      } 
      else if (ready) clearScreen(g) ;
   }
   private void drawLoadMsg(Graphics g) {
      Font msgFont=new Font("Dialog", Font.BOLD, 12) ;
      int hMargin=20, vMargin=10 ;
      
      FontMetrics fm = g.getFontMetrics(msgFont) ;
      int boxWidth = fm.stringWidth(LOAD_MSG) + (2*hMargin) ;
      int boxHeight = fm.getAscent() + fm.getDescent() + (2*vMargin) ;
      int boxLeft = (appWidth - boxWidth) / 2 ;
      int boxTop = (appHeight - boxHeight) / 2 ;
      // background
      g.setColor(BG_COLOR) ;
      g.fillRect(0,0,appWidth,appHeight) ;
      // message box
      g.setColor(Color.black) ;
      g.fillRect(boxLeft,boxTop,boxWidth+1,boxHeight+1) ;
      // message box outline
      g.setColor(Color.red) ;
      g.drawRect(boxLeft,boxTop,boxWidth,boxHeight) ;
      // message
      g.setFont(msgFont) ;
      g.setColor(Color.white) ;
      g.drawString(LOAD_MSG,boxLeft+hMargin,boxTop+vMargin+fm.getAscent()) ;
   }
   private void clearScreen(Graphics g) {
      g.setColor(BG_COLOR) ;
      g.fillRect(0,0,appWidth,appHeight) ;
   }
   void setReady() { ready = true ; }
}      
/******** APP MAIN ********
	main class: initializes applet panel, starts games,
   handles control panel & board canvas events
*/	
final class AppMain implements Runnable {
//final class AppMain implements Runnable {
   static final String DF_BOARD_FILE ="board.gif",    // default image files
                       DF_PIECES_FILE="pieces.gif",
                       DF_TITLE      ="English 16",   // default text strings
                       DF_RESTART    ="RESTART",
                       DF_UNDO       ="UNDO",
                       DF_MOVES      ="Moves",
                       DF_SOLVE      ="SOLVE",
                       DF_READY      ="Ready !",
                       DF_WIN        ="Finished !",
                       DF_LOSE       ="Game Over !",
                       DF_PERFECT    ="Congratulations !",
                       DF_SEARCHING  ="Searching ..",
                       DF_FOUND      ="Solution Found.",
                       DF_SOLVING    ="Solving ..",
                       DF_SOLVED     ="Solved !" ;
   static final int MIN_MOVES=46, MAX_MOVES=99, DELAY=1500 ;
   protected String BOARD_FILE, PIECES_FILE, TITLE, RESTART, UNDO, MOVES, SOLVE,
                  READY, WIN, LOSE, PERFECT, SEARCHING, FOUND, SOLVING, SOLVED ;

   private int bcWidth, bcHeight, moveCount ;
   private boolean gameOver ;
   private Board bd ;
   private Solver loyd ;
   private BoardCanvas bc ;
   private ControlPanel cp ;
   private Image boardImage, piecesImage ;
   private Thread solveThread ;
   private English16 app ;
   
   // constructor
   AppMain(English16 app) { 
      this.app = app ;
   }
   void initialize() {

      // * Get applet parameters *

      // image filenames
      BOARD_FILE = app.getParameter("boardFile") ; if (BOARD_FILE==null) BOARD_FILE = new String(DF_BOARD_FILE) ;
      PIECES_FILE = app.getParameter("piecesFile") ; if (PIECES_FILE==null) PIECES_FILE = new String(DF_PIECES_FILE) ;
      // text
      TITLE = app.getParameter("title") ; if (TITLE==null) TITLE = new String(DF_TITLE) ;
      RESTART = app.getParameter("restart") ; if (RESTART==null) RESTART = new String(DF_RESTART) ;
      UNDO = app.getParameter("undo") ; if (UNDO==null) UNDO = new String(DF_UNDO) ;
      MOVES = app.getParameter("moves") ; if (MOVES==null) MOVES = new String(DF_MOVES) ;
      SOLVE = app.getParameter("solve") ; if (SOLVE==null) SOLVE = new String(DF_SOLVE) ;
      READY = app.getParameter("ready") ; if (READY==null) READY = new String(DF_READY) ;
      WIN = app.getParameter("win") ; if (WIN==null) WIN = new String(DF_WIN) ;
      LOSE = app.getParameter("lose") ; if (LOSE==null) LOSE = new String(DF_LOSE) ;
      PERFECT = app.getParameter("perfect") ; if (PERFECT==null) PERFECT = new String(DF_PERFECT) ;
      SEARCHING = app.getParameter("searching") ; if (SEARCHING==null) SEARCHING = new String(DF_SEARCHING) ;
      FOUND = app.getParameter("found") ; if (FOUND==null) FOUND = new String(DF_FOUND) ;
      SOLVING = app.getParameter("solving") ; if (SOLVING==null) SOLVING = new String(DF_SOLVING) ;
      SOLVED = app.getParameter("solved") ; if (SOLVED==null) SOLVED = new String(DF_SOLVED) ;
      
      // load images
      MediaTracker tracker = new MediaTracker(app) ;
      URL url = app.getCodeBase() ;
      boardImage = app.getImage(url,BOARD_FILE) ;
      piecesImage = app.getImage(url,PIECES_FILE) ;
      tracker.addImage(boardImage,0) ;
      tracker.addImage(piecesImage,1) ;
      try { tracker.waitForAll(); }
      catch (InterruptedException e) {}

      // * Initialize applet layout *
      
      app.setBackground(app.BG_COLOR) ;
      app.setLayout(new FlowLayout(FlowLayout.CENTER,20,10)) ;
      // Board canvas
      bcWidth = boardImage.getWidth(app) ;
      bcHeight = boardImage.getHeight(app) ;
      bc = new BoardCanvas(bcWidth,bcHeight,boardImage,piecesImage,this,app) ;
      bc.resize(bcWidth,bcHeight) ;
      // Control panel
      cp = new ControlPanel(this) ;
      // add all components to enclosing panels
      Panel p1 = new Panel() ;
      p1.setBackground(Color.lightGray) ;
      p1.setLayout(new FlowLayout(FlowLayout.CENTER,5,5)) ;
      Panel p2 = new Panel() ;
      p2.setBackground(Color.gray) ;
      p2.setLayout(new FlowLayout(FlowLayout.CENTER,2,2)) ;
      p1.add(cp) ;
      p2.add(p1) ;
      app.add(bc) ;
      app.add(p2) ;
      app.validate() ;

      app.setReady() ;
      app.repaint() ;
      newGame() ;
   }
   // Start a new game
   void newGame() {
      stop() ;
      System.gc() ;
      bd = new Board(this) ;
      loyd = new Solver(bd,bc,cp,this) ;
      gameOver = false ; 
      cp.setMoveCount(moveCount=0) ;
      cp.setUndoEnable(false) ;
      cp.setSolveEnable(true) ;
      cp.setStatus(READY) ;
      bc.drawBoard(bd) ;
      bc.requestFocus() ;
   }      
   // checks for end of game conditions
   private boolean isGameOver() {
      if (bd.isEnd()) { 
         // "Perfect" if minimum moves made
         cp.setStatus(moveCount<=MIN_MOVES? PERFECT : WIN) ;
         return true ;
      }
      else if (moveCount==MAX_MOVES) {
         // maximum moves reached
         cp.setStatus(LOSE) ;
         return true ;
      }
      cp.setStatus(null) ;
      return false ;
   }
   // Board canvas mouse event
   void handleMouseDown(int sq) {
      if (!gameOver && bd.playerMove(sq)) {
         cp.setMoveCount(++moveCount) ;
         gameOver = isGameOver() ;
         cp.setUndoEnable(true) ;
         cp.setSolveEnable(false) ;
         bc.drawBoard(bd) ;
      }
   }
   // Undo event
   void handleUndo() {
      if (moveCount>0) {
         bd.undoMove() ;
         cp.setMoveCount(--moveCount) ;
         gameOver = false ;
         // back at the start?
         if (moveCount==0) {
            cp.setUndoEnable(false) ;
            cp.setSolveEnable(true) ;
            cp.setStatus(READY) ;
         }
         else cp.setStatus(null) ;
         bc.drawBoard(bd) ;
         bc.requestFocus() ;
      }
   }
   // spawn Solve thread
   public void startSolveThread() {
      if (solveThread == null) {
         solveThread = new Thread(this) ;
         solveThread.start() ;
      }
   }
   // kill thread
   public void stop() {
      if (solveThread != null) {
         solveThread.stop() ;
         solveThread = null ;
      }
   }
   // run thread
   public void run() {
      gameOver = true ;
      cp.setUndoEnable(false) ;
      cp.setSolveEnable(false) ;
      cp.setStatus(SEARCHING) ;
      // search for solution
      if (loyd.solve()) {
         cp.setStatus(FOUND) ;
         sleep(DELAY) ;
         // show solution
         cp.setStatus(SOLVING) ;
         loyd.runSolution() ;
         cp.setStatus(SOLVED) ;
      }
   }
   // thread delay
   void sleep(int ms) {
      try { Thread.currentThread().sleep(ms) ; } 
      catch (InterruptedException e) {}
   }
}      
/******** BOARD ********
   maintains board status, makes moves
   Board layout:  | 0| 1| 2|
                  | 3| 4| 5|
                  | 6| 7| 8| 9|10|
                        |11|12|13|
                        |14|15|16|
*/   
final class Board {
   static final int BD_SIZE=17, START_TARGET=8,EMPTY=0, LIGHT=1, DARK=2,
      startBoard[]= {LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,
                     EMPTY,
                     DARK,DARK,DARK,DARK,DARK,DARK,DARK,DARK},
      endBoard[]  = {DARK,DARK,DARK,DARK,DARK,DARK,DARK,DARK,
                     EMPTY,
                     LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT,LIGHT},
      legal[][] = { { 2,6,1,3},{ 7,0,2,4},{ 0,8,1,5},
                    { 5,0,4,6},{ 1,3,5,7},{ 3,11,2,4,8},
                    { 0,8,3,7},{ 1,9,4,6,8},
                    { 2,10,14,6,5,7,9,11},
                    { 7,15,8,10,12},{ 8,16,9,13},
                    { 5,13,8,12,14},{ 9,11,13,15},{11,10,12,16},
                    { 8,16,11,15},{ 9,12,14,16},{10,14,13,15} } ;
   
   private int board[]=new int[BD_SIZE], history[], histPtr, target ;
   private AppMain main ;
   
   // constructor
   Board(AppMain main) {
      this.main = main ;
      // initalize board
      for (int i=0; i<BD_SIZE; i++) board[i] = startBoard[i] ;
      // target is always current empty square
      target = START_TARGET ;
      // initialize history
      history = new int[main.MAX_MOVES+1] ;
      history[0] = target ;
      histPtr = 1 ;
   }
   int getSquare(int sq) { return board[sq] ; }
   void setSquare(int sq,int val) { board[sq] = val ; }

   // moves piece and returns new target value
   int movePiece(int source, int target) {
      board[target] = board[source] ;
      board[source] = EMPTY ;
      return source ;
   }
   // makes player move
   boolean playerMove(int source) {
      if (isLegalMove(source,target)) {
         // make move
         target = movePiece(source,target) ;
         // save it
         history[histPtr++] = source ;
         return true ;
      }
      return false ;
   }
   // returns true if final position reached
   boolean isEnd() {
      for (int i=0; i<BD_SIZE; i++) 
         if (board[i]!=endBoard[i]) return false ;
      return true ;
   }
   // resets board to previous move using history data
   void undoMove() {
      if (histPtr>1) {
         // get last saved move
         target = history[--histPtr] ;
         // make it
         target = movePiece(history[histPtr-1],target) ;
      }
   }
   private boolean isLegalMove(int source,int target) {
      for (int i=0; i<legal[target].length; i++)
         if (legal[target][i]==source) return true ;
      return false ;
   }
}         
/******** SOLVER ********
   searches for solution, runs found solution
*/   
final class Solver {
   static final int SOLVE_DELAY=800 ;
   private boolean isSolved ;
   private int solution[], moves, nodes, searchDelay=120 ;
   private Board bd ;
   private ControlPanel cp ;
   private BoardCanvas bc ;
   private AppMain main ;
   
   // constructor
   Solver(Board bd,BoardCanvas bc,ControlPanel cp,AppMain main) {
      this.bd = bd ;
      this.bc = bc ;
      this.cp = cp ;
      this.main = main ;
   }
   // show the solution we've found
   void runSolution() {
      if (solution!=null) {
         // reset board         
         for (int i=0; i<bd.BD_SIZE; i++) bd.setSquare(i,bd.startBoard[i]) ;
         bc.drawBoard(bd) ;
         // as with Board class, target is current empty square
         int target = bd.START_TARGET ;
         // move animation loop
         for (int i=0; i<moves; i++) {
            // delay            
            main.sleep(SOLVE_DELAY) ;
            // move piece and update counter
            target = bd.movePiece(solution[i],target) ;
            cp.setMoveCount(i+1) ;
            // redraw board
            bc.drawBoard(bd) ;
         }
      }
   }
   // initialize search
   boolean solve() {      
      moves = nodes = 0 ;
      isSolved = false ;
      solution = new int[main.MAX_MOVES+1] ;
         
      search(bd.START_TARGET) ;
      return isSolved ;
   }      
   // recursive DFS routine
   void search(int target) {
      nodes++ ;
      // base to end recursion
      if (bd.isEnd()) {
         isSolved = true ;
         return ;
      }
      // generate move list
      int mv[] = genMoves(target) ;
      if (mv==null) return ;
      
      // loop through move list
      for (int i=0; i<mv.length; i++) {
         // save move
         solution[moves++] = mv[i] ;
         // make move (do not update target!)
         bd.movePiece(mv[i],target) ;
         // draw board with animation delay
         bc.drawBoard(bd) ;
         main.sleep(searchDelay) ;
         // change delay if needed
         if (nodes>1000) searchDelay = 5 ;
         else if (nodes>500) searchDelay = 10 ;
         else if (nodes>200) searchDelay = 30 ;
         else if (nodes>100) searchDelay = 60 ;
         else if (nodes>50) searchDelay = 90 ;
         // recursive call
         search(mv[i]) ;
         // quit if done
         if (isSolved) return ;
         else {
            // unmake move
            bd.movePiece(target,mv[i]) ;
            moves-- ;
         }
      }
   }
   // generate all legal moves where pieces move forward
   // (toward opposite end of board)
   int[] genMoves(int target) {
      int list1[] = new int[8], moveCount=0 ;
      
      for (int i=0; i<bd.legal[target].length; i++) {
         // dark pieces
         if ( (bd.legal[target][i]>target) && (bd.getSquare(bd.legal[target][i])==bd.DARK) )
            list1[moveCount++] = bd.legal[target][i] ;
         // light pieces
         else if ( (bd.legal[target][i]<target) && (bd.getSquare(bd.legal[target][i])==bd.LIGHT) )
            list1[moveCount++] = bd.legal[target][i] ;
      }
      if (moveCount!=0) {
         int list2[] = new int[moveCount] ;
         int wLeft = moveCount ;
         // shuffle movelist into correctly sized array
         // (we want to search the tree randomly)
         for (int i=0; i<moveCount; i++) {
              int rnd = randomInt(wLeft) ;
              list2[i] = list1[rnd];
              list1[rnd] = list1[--wLeft];
         }
         return list2 ;
      }
      return null ;
   }
   int randomInt(int n) { return (int)(n*Math.random()) ; }
}
/******** BOARD CANVAS ********
   controls board graphics, 
   passes mouse events to main class
*/   
final class BoardCanvas extends Canvas {
   final static int squareCoord[][]= { {55,109},{82,82},{110,54},{82,137},{110,109},{137,82},
                                       {110,164},{137,137},{165,109},{192,82},{220,54},{192,137},
                                       {220,109},{247,82},{220,164},{247,137},{275,109}, } ;
   private int bcWidth, bcHeight, pieceWidth, pieceHeight, pieceWidthOffset, pieceHeightOffset ;
   private Image bufferImage, boardImage, piecesImage, lightPieceImage, darkPieceImage ;
   private MediaTracker tracker ;
   private Graphics buffer ;
   private AppMain main ;
   private English16 app ;

   BoardCanvas(int bcWidth,int bcHeight,Image boardImage,Image piecesImage,
               AppMain main,English16 app) { 
      this.bcWidth = bcWidth ;
      this.bcHeight = bcHeight ;
      this.boardImage = boardImage ;
      this.piecesImage = piecesImage ;
      this.main = main ;
      this.app = app ;

      pieceWidth = piecesImage.getWidth(app)/2 ;
      pieceHeight = piecesImage.getHeight(app) ;
      pieceWidthOffset = pieceWidth/2 ;
      pieceHeightOffset = pieceHeight/2 ;
      // extract piece images
      tracker = new MediaTracker(app) ;
      lightPieceImage = extractImage(0,0,pieceWidth,pieceHeight) ;
      darkPieceImage = extractImage(pieceWidth,0,pieceWidth,pieceHeight) ;
      try { tracker.waitForAll() ; }
      catch (InterruptedException e) {}
   }
   void drawBoard(Board bd) {
      if (buffer==null) {
         bufferImage = createImage(bcWidth,bcHeight) ;
         buffer = bufferImage.getGraphics() ;
      }
      // draw the board and pieces
      buffer.drawImage(boardImage,0,0,this) ;
      for (int i=0; i<bd.BD_SIZE; i++) {
         if (bd.getSquare(i)==bd.LIGHT) 
            buffer.drawImage(lightPieceImage,
               squareCoord[i][0]-pieceWidthOffset,
               squareCoord[i][1]-pieceHeightOffset,this) ;
         else if (bd.getSquare(i)==bd.DARK) 
            buffer.drawImage(darkPieceImage,
               squareCoord[i][0]-pieceWidthOffset,
               squareCoord[i][1]-pieceHeightOffset,this) ;
      }
      repaint() ;
   }
   public void paint(Graphics g) { update(g) ; }
   public void update(Graphics g) { g.drawImage(bufferImage,0,0,this) ; }
   
   // mouse event handler
   public boolean mouseDown(Event e,int x,int y) {   
      int sq = pixelToSquare(x,y) ;
      if (sq!=-1) main.handleMouseDown(sq) ;
      return true ; 
   }
   // pixel-square conversion
   int pixelToSquare(int x, int y) {
      for (int i=0; i<squareCoord.length; i++) {
         if ( x > (squareCoord[i][0]-pieceWidthOffset) && 
              x < (squareCoord[i][0]+pieceWidthOffset) &&
              y > (squareCoord[i][1]-pieceHeightOffset) &&
              y < (squareCoord[i][1]+pieceHeightOffset) )
            return i ;
      }
      return -1 ;
   }
   // Extracts piece images
   private Image extractImage(int x,int y,int w,int h) {
      Image newImage ;
      ImageFilter filter;
      ImageProducer producer;

      filter = new CropImageFilter(x,y,w,h);
      producer = new FilteredImageSource(piecesImage.getSource(),filter);
      newImage = app.createImage(producer);
      tracker.addImage(newImage,0);
      return newImage ;
   }
}
/******** CONTROL PANEL ********
   the UI components for the game
*/   
final class ControlPanel extends Panel {
   static final Font titleFont =new Font("SansSerif", Font.BOLD, 12),
                     textFont  =new Font("SansSerif", Font.PLAIN, 12),
                     smallFont =new Font("SansSerif", Font.PLAIN, 10),
                     movesFont =new Font("Courier", Font.BOLD, 12),
                     statusFont=new Font("Dialog", Font.PLAIN, 11) ;
   private Button bRestart, bUndo, bSolve ;
   private TextField tfMoveCount, tfStatus ;
   private Label lTitle, lMoves ;
   private AppMain main ;

   // constructor
   ControlPanel(AppMain main) {
      this.main = main ;

      setLayout(new GridLayout(0,1,0,5)) ;
      setBackground(Color.lightGray) ;
      setFont(textFont) ;
      // Title
      add(lTitle=new Label(main.TITLE,Label.CENTER)) ;
      lTitle.setFont(titleFont) ;
      // Restart and Undo Buttons
      add(bRestart=new Button(main.RESTART)) ;
      add(bUndo=new Button(main.UNDO)) ;
      // MoveCount panel
      Panel mcp = new Panel() ;
      mcp.setLayout(new GridLayout(0,2,5,0)) ;
      mcp.add(tfMoveCount=new TextField(3)) ;      
      mcp.add(lMoves=new Label(main.MOVES,Label.LEFT)) ;
      lMoves.setFont(smallFont) ;
      add(mcp) ;
      tfMoveCount.setFont(movesFont) ;
      tfMoveCount.setForeground(Color.black) ;
      tfMoveCount.setBackground(Color.lightGray) ;
      tfMoveCount.setEditable(false) ;
      // Status textfield
      add(tfStatus=new TextField(12)) ;
      tfStatus.setFont(statusFont) ;
      tfStatus.setForeground(Color.white) ;
      tfStatus.setBackground(Color.black) ;
      tfStatus.setEditable(false) ;
      // Solve Button
      add(bSolve=new Button(main.SOLVE)) ;
      validate() ;
   }
   // Button Events
   public boolean action(Event e,Object o) { 
      if (main.RESTART.equals(o)) main.newGame() ;
      else if (main.UNDO.equals(o)) main.handleUndo() ;
      else if (main.SOLVE.equals(o)) main.startSolveThread() ;
      return true ;
   }
   // Button enable/disable
   void setUndoEnable(boolean b) { bUndo.enable(b) ; }
   void setSolveEnable(boolean b) { bSolve.enable(b) ; }
   
   // Textfield updates
   void setMoveCount(int i) { 
      String mc = Integer.toString(i) ;
      tfMoveCount.setText(mc.length()==1?"  "+mc:" "+mc) ;
   }
   void setStatus(String st) { 
      tfStatus.setText(st==null?" ":" "+st) ; 
   }
}